#include <stdint.h>
#include "fip_urv_regs.h"
#include "fip_urv_cmd.h"

/* I2C open-core registers.  */
#define OC_I2C_PRES_LO 0
#define OC_I2C_PRES_HI 1
#define OC_I2C_CTRL    2
#define OC_I2C_RXTX    3
#define OC_I2C_STAT    4
#define OC_I2C_CR      4
#define OC_I2C_CFG     5

static unsigned extract_word (volatile uint32_t *buf)
{
  unsigned res;

  /* Little endian.  */
  res = (buf[0] & 0xff) << 0;
  res |= (buf[1] & 0xff) << 8;
  res |= (buf[2] & 0xff) << 16;
  res |= (buf[3] & 0xff) << 24;

  return res;
}

static void insert_word(unsigned char *buf, unsigned val)
{
  buf[0] = (val >> 0) & 0xff;
  buf[1] = (val >> 8) & 0xff;
  buf[2] = (val >> 16) & 0xff;
  buf[3] = (val >> 24) & 0xff;
}

struct i2c_cmd {
  unsigned char state;
  unsigned char len;
  unsigned char coff;
  unsigned char boff;
  unsigned cmds[16];
  unsigned buf[32];
};

static inline void
do_i2c_write (struct i2c_cmd *i2c)
{
  volatile struct fip_urv_regs *regs =
    (volatile struct fip_urv_regs *)0x100000;
  unsigned char stop =
    i2c->len == 1 && i2c->state == I2C_CMD_WRITE_STOP;
  regs->i2c[OC_I2C_RXTX] = i2c->cmds[i2c->coff];
  regs->i2c[OC_I2C_CR] = 0x10 | (stop ? 0x40 : 0);  /* WRITE */
  i2c->coff++;
}

static inline void
do_i2c_read (struct i2c_cmd *i2c)
{
  volatile struct fip_urv_regs *regs =
    (volatile struct fip_urv_regs *)0x100000;

  unsigned char stop =
    i2c->len == 1 && i2c->state == I2C_CMD_READ_STOP;
  regs->i2c[OC_I2C_CR] = 0x20 | (stop ? 0x48 : 0);  /* READ */
}

static void
do_i2c_done (struct i2c_cmd *i2c, unsigned char stat)
{
  volatile struct fip_urv_regs *regs =
    (volatile struct fip_urv_regs *)0x100000;

  /* Save values in case of error.  */
  i2c->buf[31] = stat;
  i2c->buf[30] = i2c->state;

  if (i2c->state == I2C_CMD_START) {
    if (stat == 0x41) {
      i2c->state = I2C_CMD_RUN;
      return;
    }
  }
  else if (i2c->state == I2C_CMD_WRITE
           || i2c->state == I2C_CMD_WRITE_STOP) {
    if (stat == 0x41 || (stat == 0x01 && i2c->len == 1)) {
      i2c->len--;
      if (i2c->len == 0)
        i2c->state = I2C_CMD_RUN;
      else
        do_i2c_write(i2c);
      return;
    }
  }
  else if (i2c->state == I2C_CMD_READ
           || i2c->state == I2C_CMD_READ_STOP) {
    /* We are the master.  No need to check STAT, as the ACK/BUSY
       bits are in fact set by us.  */
    i2c->buf[i2c->boff] = regs->i2c[OC_I2C_RXTX];
    i2c->boff++;
    i2c->len--;
    if (i2c->len == 0)
      i2c->state = I2C_CMD_RUN;
    else
      do_i2c_read (i2c);
    return;
  }
  else if (i2c->state == I2C_CMD_READLEN) {
    if (stat == 0x41) {
      i2c->len = regs->i2c[OC_I2C_RXTX];
      i2c->buf[i2c->boff] = i2c->len;
      i2c->boff++;
      i2c->state = I2C_CMD_RUN;
      return;
    }
  }

  i2c->state = I2C_CMD_ERR;
}

static void
do_i2c_run (struct i2c_cmd *i2c)
{
  volatile struct fip_urv_regs *regs =
    (volatile struct fip_urv_regs *)0x100000;

  unsigned char cmd = i2c->cmds[i2c->coff];

  /* Default state: the command.  */
  i2c->state = cmd & 0x0f;

  if (cmd & 0xf0)
    i2c->len = cmd >> 4;

  if (cmd == I2C_CMD_IDLE) {
    /* Last command.  */
  }
  else if (cmd == I2C_CMD_START) {
    regs->i2c[OC_I2C_RXTX] = i2c->cmds[i2c->coff + 1];
    regs->i2c[OC_I2C_CR] = 0x90;  /* START + WRITE */
    i2c->coff += 2;
  }
  else if ((cmd & 0x0f) == I2C_CMD_WRITE
           || (cmd & 0x0f) == I2C_CMD_WRITE_STOP) {
    i2c->coff++;
    do_i2c_write(i2c);
  }
  else if ((cmd & 0x0f) == I2C_CMD_READ
           || (cmd & 0x0f) == I2C_CMD_READ_STOP
           || (cmd & 0x0f) == I2C_CMD_READLEN) {
    i2c->coff++;
    do_i2c_read(i2c);
  }
  else {
    /* Bad command.  */
    i2c->state = I2C_CMD_ERR;
  }
}

int
main (void)
{
  volatile struct fip_urv_regs *regs =
    (volatile struct fip_urv_regs *)0x100000;
  struct i2c_cmd i2c;
  int j;
  unsigned char reply[8];
  unsigned char rdy;
  unsigned int leds;

  /* Init.  */
  leds = 1;
  j = 0;
  rdy = 0;
  i2c.state = I2C_CMD_IDLE;
  i2c.coff = 0;
  i2c.boff = 0;
  i2c.len = 0;

  /* Init i2c: prescaler (400khz) */
  regs->i2c[OC_I2C_PRES_LO] = 25;
  regs->i2c[OC_I2C_PRES_HI] = 0;
  regs->i2c[OC_I2C_CTRL] = 0x80;  /* Enable */

  /* Infinite loop.  */
  while (1)
    {
      /* Check for FIP message.  */
      unsigned int status = regs->fip_status;

      /* Control leds.  */
      if (leds & 0x80)
        regs->leds = leds;
      else
        {
          regs->leds = (leds << 2) | status;
          j++;
          if (j == 1000000)
            {
              leds = (leds >> 1) | ((leds & 1) << 3);
              j = 0;
            }
        }


      /* Send leds status.  */
      if (status & 1)
        {
          /* Received.  */
          unsigned char cmd;

          /* Consume variable.  */
          regs->fip_var1 = 1;
          cmd = regs->fip_reg[0x02];

          /* Default reply: command + id.  */
          reply[0] = cmd;
          reply[1] = regs->fip_reg[0x03];

          /* No switch (no .rodata)  */
          if (cmd == CMD_SET_LEDS) {
              leds = regs->fip_reg[0x06];
          }
          else if (cmd == CMD_READ_WORD) {
            unsigned *addr;

            addr = (unsigned *)extract_word(&regs->fip_reg[0x6]);
            insert_word(&reply[4], *addr);
          }
          else if (cmd == CMD_WRITE_BLOCK || cmd == CMD_WRITE_WORD) {
            unsigned *addr;
            unsigned val;
            unsigned k;

            addr = (unsigned *)extract_word(&regs->fip_reg[0x6]);
            for (k = 0; k < cmd; k++)
              {
                val = extract_word(&regs->fip_reg[0xa + 4*k]);
                addr[k] = val;
              }
          }
          else if (cmd == CMD_I2C_TFR) {
            if ((i2c.state & 0x0f) != I2C_CMD_IDLE) {
              /* Busy */
              reply[4] = 1;
            }
            else {
              unsigned k;
              unsigned clen = regs->fip_reg[0x6];
              for (k = 0; k < clen; k++)
                i2c.cmds[k] = regs->fip_reg[0x7 + k];
              i2c.cmds[clen] = I2C_CMD_IDLE;
              i2c.coff = 0;
              i2c.boff = 0;
              i2c.state = I2C_CMD_RUN;
              reply[4] = 0;
            }
          }
          else if (cmd == CMD_I2C_STAT) {
            reply[4] = i2c.state;
            reply[5] = i2c.len;
            reply[6] = i2c.coff;
            reply[7] = i2c.boff;
          }
          else if (cmd == CMD_I2C_BUF) {
            unsigned off = regs->fip_reg[0x6];
            reply[4] = i2c.buf[off + 0];
            reply[5] = i2c.buf[off + 1];
            reply[6] = i2c.buf[off + 2];
            reply[7] = i2c.buf[off + 3];
          }
          else {
            reply[0] = CMD_ERROR;
          }

          /* Ready to send a reply.  */
          rdy = 1;
          /* Variable has been consumed.  */
          regs->fip_var1 = 0;
        }

      if ((status & 2) && rdy)
        {
          int i;

          /* Produce variable.  */
          regs->fip_var3 = 1;

          for (i = 0; i < sizeof(reply); i++)
            regs->fip_reg[0x102 + i] = reply[i];

          regs->fip_var3 = 0;

          rdy = 0;
        }

      if ((i2c.state & 0x0f) != I2C_CMD_IDLE) {
        if (i2c.state != I2C_CMD_RUN) {
          /* For end of transfer.  */
          unsigned char stat = regs->i2c[OC_I2C_STAT];
          if (!(stat & 2))
            do_i2c_done (&i2c, stat);
        }

        if (i2c.state == I2C_CMD_RUN)
          do_i2c_run (&i2c);
      }
    }
}
